from multi_time_series_connectedness.connectedness import Connectedness
import pandas as pd
import pickle
import sys


class RollingConnectedness:
    def __init__(self, data, max_lag, data_periods, start_at, end_at):
        self.data = data
        self.max_lag = max_lag
        self.data_periods = data_periods
        self.name = [col for col in data.columns if col != 'time'] + ['all']
        self.start_at = start_at
        self.end_at = end_at

        # save the calculated connectedness
        self.split_data = None
        self.rolling_connectedness = None
        self.accuracy_list = None

    def divide_timeseries_volatilities(self):
        data_eariler_than_start_at = self.data[self.data['time'] <= self.start_at]
        if (len(data_eariler_than_start_at) < self.data_periods):
            print("There is not enough data for the first connectedness")
            sys.exit()

        filtered_time = self.data[(self.data['time'] > self.start_at) & (self.data['time'] <= self.end_at)]['time']
        print(filtered_time)
        split_data = {}
        for index, time in filtered_time.items():
            last_rows = self.data.iloc[index-self.data_periods - 1:index - 1] # I think this line will cause problem
            split_data[time] = last_rows
        self.split_data = split_data

    def calculate(self, store_result_at, callback_after_one_connectedness=None):
        self.divide_timeseries_volatilities()
        restructured_connectedness_timeseries = pd.DataFrame()
        for key, data in self.split_data.items():
            start_date = data["time"].iloc[0]
            end_date = data["time"].iloc[self.data_periods-1]
            period = start_date + " ~ " + end_date
            print("calculate connectedness for period, %s with data between %s"
                  % (end_date, period))

            conn = Connectedness(data)
            conn.calculate_full_connectedness()
            conn.rename_table(self.name)
            conn.flatten_connectedness()

            restructured_connectedness = conn.restructure_connectedness
            restructured_connectedness['period'] = key
            restructured_connectedness.set_index('period', inplace=True)
            if callback_after_one_connectedness:
                callback_after_one_connectedness(restructured_connectedness)
            restructured_connectedness_timeseries = pd.concat(
                [restructured_connectedness_timeseries, restructured_connectedness], 
                ignore_index=False
            )
        restructured_connectedness_timeseries['forecast_at'] = restructured_connectedness_timeseries.index.to_series().shift(-conn.forecast_at_next_period)
        print(restructured_connectedness_timeseries)
        self.rolling_connectedness = restructured_connectedness_timeseries

        with open(store_result_at, 'wb') as f:
            pickle.dump(self.rolling_connectedness, f)

    def plot_rolling():
        pass



import pickle
import pandas as pd
import numpy as np
import os


class Volatility:
    def __init__(self, n=2):
        self.n = n

    def yang_zhang_volatility(self, data, name):
        """
        :param data: a list with Open, High, Low, Close price
        :param name: the name of the volatility column
        :return: A DataFrame with time and volatility data
        """
        # define required variables
        o_c = (data['Open'] / data['Close'].shift(1)).apply(np.log)
        c_o = (data['Close'] / data['Open']).apply(np.log)
        h_o = (data['High'] / data['Open']).apply(np.log)
        l_o = (data['Low'] / data['Open']).apply(np.log)

        # overnight volatility
        vo = o_c.rolling(window=self.n).apply(np.var, raw=True)

        # today(open to close) volatility
        vt = c_o.rolling(window=self.n).apply(np.var, raw=True)

        # rogers-satchell volatility
        rs_fomula = h_o * (h_o - c_o) + l_o * (l_o - c_o)
        rs = rs_fomula.rolling(window=self.n, center=False).sum() * (1.0 / self.n)

        # super parameter
        k = 0.34 / (1 + (self.n + 1) / (self.n - 1))

        # yang-zhang
        result = (vo + k * vt + (1 - k) * rs).apply(np.sqrt)

        result_df = result.to_frame(name=name)

        return pd.concat([data['time'], result_df], axis=1)

    def price_data_to_volatility(self, datasets):
        volatilities = None
        for key, value in datasets.items():
            volatility = self.yang_zhang_volatility(value, key)
            if volatilities is None:
                volatilities = volatility
            else:
                volatilities = volatilities.merge(volatility, on='time', how='outer')

        return volatilities

    def calculate(self, directory, start_at, end_at, save_path=None):
        datasets = {}
        for filename in os.listdir(directory):
            if filename.endswith('.csv'):
                data = pd.read_csv(os.path.join(directory, filename))
                filtered_data = data[(data['time'] >= start_at) & (data['time'] <= end_at)]
                datasets[filename] = filtered_data
        volatilities = self.price_data_to_volatility(datasets)

        with open(save_path, 'wb') as f:
            pickle.dump(volatilities, f)



# import required modules
import numpy as np
import math
from sklearn import linear_model
import operator
# ======================


def var_y(data, lag=1):
    """
    :param data: volatiltiies dataframe with Date
    :param lag: the lags of AR
    :return: for example, if lag = 4, then the first part of 4 periods would be
    # deleted because we cannot estimate them
    """
    y = data.drop("time", axis=1).drop(data.index[:lag]).values.transpose()
    return y


def shift_right(data, n_shift):
    """
    :param data: matrix without Date
    :param n_shift: the number of lags
    :return: shifted data (lag is the number of position to shift and filled
    # the blank with NA)
    """
    nrow = data.shape[0]
    ncol = data.shape[1]
    matrix_na = np.empty((nrow, n_shift,))
    matrix_na[:] = np.nan
    matrix_num = np.delete(data, [list(range(ncol - n_shift, ncol))], 1)
    result = np.concatenate((matrix_na, matrix_num), axis=1)
    return result


def var_x(data, lag=1):
    """
    :param data: volatiltiies dataframe with Date
    :param lag: the lag od AR
    :return: a matrix with stacked shifted past period data
    """
    y = data.drop("time", axis=1).values.transpose()
    x = y
    if lag > 1:
        for i in range(1, lag):
            x_down = shift_right(y, i)
            x = np.concatenate((x, x_down), 0)
        x = np.delete(x, range(lag - 1), 1)
    x = np.delete(x, x.shape[1] - 1, 1)
    return x


def ols(sy, sx):
    """
    :param sy: output of var_y
    :param sx: output of var_x
    :return: ols estimated coef
    """
    a = np.dot(sy, sx.transpose())
    b = np.linalg.inv(np.dot(sx, sx.transpose()))
    coef_result = np.dot(a, b)
    return coef_result


def mle_sigma(sy, sx, coef):
    t = sx.shape[1]
    std = sy - np.dot(coef, sx)
    mle_sigma_result = np.dot(std, std.transpose()) / t
    return mle_sigma_result


def aic(mle_sigma_estimates, coef, t):
    length = coef.shape[0] * coef.shape[1]
    eq_0 = np.count_nonzero(coef == 0)
    length = length - eq_0
    # if the size of data is too small, then the det may be 0, meaning it can not estimate it
    aic_result = math.log(np.linalg.det(mle_sigma_estimates)) + 2 / t * length
    return aic_result


def lag_chooser(data, max_lag):
    list_aic = []
    for i in range(1, max_lag + 1):
        sx = var_x(data, i)
        sy = var_y(data, i)
        t = sx.shape[1]
        reg = linear_model.LinearRegression(fit_intercept=False)
        reg.fit(sx.transpose(), sy.transpose())
        coef = reg.coef_
        mle_sigma_result = mle_sigma(sy, sx, coef)
        aic_result = aic(mle_sigma_result, coef, t)
        list_aic.append(aic_result)
    index, value = min(enumerate(list_aic), key=operator.itemgetter(1))
    return index + 1, value


class Coef:

    def __init__(self, data, max_lag):
        # The variables we need to launch this class
        # the lag chooses from lag_chooser
        self.lag = lag_chooser(data, max_lag)
        # the x and y to calculate coef
        self.x = var_x(data, self.lag[0])
        self.y = var_y(data, self.lag[0])

        # The place where to save calculated coef
        self.OLS_coef = None
        self.LASSO_coef = None
        self.LASSO_score = None
        self.LASSO_alpha = None

        # The place where to save calculated sigma
        self.OLS_sigma = None

        # accuracy
        self.accuracy = None

    def f_ols_coef(self):

        sx = self.x
        sy = self.y

        reg = linear_model.LinearRegression(fit_intercept=False)
        reg.fit(sx.transpose(), sy.transpose())
        self.accuracy = reg.score(sx.transpose(), sy.transpose())
        ols_coef_result = reg.coef_
        self.OLS_coef = ols_coef_result
        self.OLS_sigma = mle_sigma(sy, sx, self.OLS_coef)

    def f_lasso_coef(self, cv_value, max_iter):

        data = self.Data
        lag = self.Lag

        sx = self.var_x(data, lag[0])
        sy = self.var_y(data, lag[0])

        lasso_model = (linear_model.
                       MultiTaskLassoCV(cv=cv_value, fit_intercept=False,
                                        max_iter=max_iter).
                       fit(sx.transpose(), sy.transpose()))

        self.LASSO_alpha = lasso_model.alpha_

        clf = (linear_model.
               MultiTaskLasso(fit_intercept=False, alpha=self.LASSO_alpha,
                              max_iter=1000))

        clf.fit(sx.transpose(), sy.transpose())

        self.LASSO_score = clf.score(sx.transpose(), sy.transpose())
        self.LASSO_coef = clf.coef_

    """
    ## Ridge
# SX = VAR_X(stock_volatility, lag[0]) # 23 determined by lag_chooser in OLS
# SY = VAR_Y(stock_volatility, lag[0]) # 23 determined by lag_chooser in OLS
# n_alphas = 200
# alphas = np.logspace(-10, 1, n_alphas)
# Ridge_model = linear_model.RidgeCV(alphas=alphas, cv=10, fit_intercept=False).fit(SX.transpose(), SY.transpose())
# alpha = Ridge_model.alpha_
# clf = linear_model.Ridge(fit_intercept=False, alpha=alpha)
# clf.fit(SX.transpose(), SY.transpose())
# clf.score(SX.transpose(), SY.transpose())
 #coef_Ridge = clf.coef_ # get it

## Elastic net (combine LASSO and ridge, l = 1 for LASSO and l = 0 for Ridge)
# SX = VAR_X(stock_volatility, lag[0]) # 23 determined by lag_chooser in OLS
# SY = VAR_Y(stock_volatility, lag[0]) # 23 determined by lag_chooser in OLS
# n_alphas = 10
# alphas = np.logspace(-10, 1, n_alphas)
# ElasticNet_model = linear_model.MultiTaskElasticNetCV(alphas=alphas, cv=10, fit_intercept=False, l1_ratio=0.5, max_iter=10000).fit(SX.transpose(), SY.transpose())
 #alpha = ElasticNet_model.alpha_
# clf = linear_model.MultiTaskElasticNet(fit_intercept=False, alpha=alpha, max_iter=10000)
# clf.fit(SX.transpose(), SY.transpose())
# clf.score(SX.transpose(), SY.transpose())
# coef_ElasticNet = clf.coef_

# get the coefficients
# os.chdir("/Users/rucachen/Desktop/open_pycharm_virtualenv_3.6.4/financial_connectedness/coef")
# dict_coef = {}
# name = ["OLS", "LASSO", "Ridge", "ElasticNet"]
# list = [coef_OLS, coef_LASSO, coef_Ridge, coef_ElasticNet]
# for i in range(len(name)):
#     dict_coef[name[i]] = list[i]

# for i in range(len(name)):
#     Name = name[i] + "_" + "coef" + ".csv"
#     coef = pd.DataFrame(dict_coef[name[i]]).to_csv(Name)
    """
